# Copyright 2016 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import functools
import os

from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import hookenv, host, unitdata
from charmhelpers.core.templating import render
from charms.layer import snap
from charms.reactive import (
    remove_state,
    set_state,
    when,
    when_any,
    when_not,
)
from charms.reactive.helpers import data_changed
from jinja2 import Template
import yaml

SNAP_NAME = "prometheus-openstack-exporter"
SVC_NAME = "snap.prometheus-openstack-exporter." "prometheus-openstack-exporter.service"
VAR_SNAP_ETC = "/var/snap/prometheus-openstack-exporter/current"
VAR_SNAP_COMMON = "/var/snap/prometheus-openstack-exporter/common"

MAIN, CREDS, CACERT = ("main", "creds", "cacert")
CONFIG_MAP = {
    "main": {
        "source": "prometheus-openstack-exporter.yaml",
        "target": "{etcdir}/prometheus-openstack-exporter.yaml",
    },
    "creds": {"source": "admin.novarc", "target": "{etcdir}/admin.novarc"},
    "cacert": {"source": "ca.crt", "target": "{etcdir}/ca.crt"},
}


def config_paths(key):
    config_path = CONFIG_MAP[key]
    config_path["target"] = config_path["target"].format(etcdir=VAR_SNAP_ETC)
    return config_path


def reconfig_on_change(key, data):
    if not data_changed(key, data):
        hookenv.log(
            "{} data unchanged, skipping reconfig".format(key), level=hookenv.DEBUG
        )
        return
    unitdata.kv().set(key, data)
    hookenv.log("{} data changed, triggering reconfig".format(key), level=hookenv.DEBUG)
    set_state("exporter.do-reconfig")


@when_not("exporter.installed")
def install_packages():
    hookenv.status_set("maintenance", "Installing software")
    config = hookenv.config()
    channel = config.get("snap_channel", "stable")
    # required for offline installs
    # uses charm store if no "core" resource is provided
    snap.install("core")
    snap.install(SNAP_NAME, channel=channel, force_dangerous=False)
    hookenv.status_set("maintenance", "Software installed")
    set_state("exporter.installed")


@when("swift-storage.available")
def save_swift_hosts(swift):
    services = swift.services()
    swift_hosts = [
        h["hostname"]
        for h in functools.reduce(lambda x, y: x + y, [s["hosts"] for s in services])
    ]
    hookenv.log("services => {}, hosts => {}".format(services, swift_hosts))
    reconfig_on_change("swift_hosts", swift_hosts)


@when("identity-credentials.connected")
def configure_keystone_username(keystone):
    username = "prometheus-openstack-exporter"
    keystone.configure(username)


@when("identity-credentials.available")
def save_creds(keystone):
    reconfig_on_change("keystone-relation-creds", keystone.get_creds())


@when_any("exporter.installed", "exporter.do-check-reconfig")
def check_reconfig_exporter():
    config = hookenv.config()
    if data_changed("exporter.config", config):
        set_state("exporter.do-reconfig")
    remove_state("exporter.do-check-reconfig")


def convert_from_base64(v):
    if not v:
        return v
    if v.startswith("-----BEGIN"):
        return v
    try:
        return base64.b64decode(v).decode()
    except TypeError:
        return v


# allow user to override credentials (and the need to be related to keystone)
# with 'os-credentials' YAML dict
def get_credentials():
    config = hookenv.config()
    config_creds_yaml = config.get("os-credentials")
    creds = {}
    if config_creds_yaml:
        config_creds = yaml.safe_load(config_creds_yaml)
        creds = {
            "credentials_username": config_creds["username"],
            "credentials_password": config_creds["password"],
            "credentials_project": config_creds.get("tenant_name", "admin"),
            "region": config_creds["region_name"],
            "auth_url": config_creds["auth_url"],
        }
        identity_api_version = config_creds.get("identity_api_version", 2)
        if identity_api_version == 3:
            creds["credentials_user_domain_name"] = config_creds["user_domain_name"]
            creds["credentials_project_name"] = config_creds.get(
                "project_name", "admin"
            )
            creds["credentials_identity_api_version"] = identity_api_version
            creds["credentials_project_domain_name"] = config_creds.get(
                "project_domain_name", "admin_domain"
            )
            creds["credentials_project_id"] = config_creds["project_id"]
            creds["credentials_project_domain_id"] = config_creds["project_domain_id"]
    else:
        kv = unitdata.kv()
        creds = kv.get("keystone-relation-creds")
        # lp#1785864 - Workaround Keystone relation not yet ready.
        if creds is None:
            hookenv.log(
                "get_credentials: config os-credentials not set and "
                "identity-credentials relation not yet ready"
            )
            return None
        # The "dict" given by the relation and the variables expected by the
        # template aren't the same, so we need to manually map some values.
        if "domain" in creds:
            if "credentials_user_domain_name" not in creds:
                creds["credentials_user_domain_name"] = creds["domain"]
            if "credentials_project_domain_name" not in creds:
                creds["credentials_project_domain_name"] = creds["domain"]
        if "api_version" in creds:
            if "identity_api_version" not in creds:
                creds["credentials_identity_api_version"] = int(creds["api_version"])
    ssl_ca = convert_from_base64(config.get("ssl_ca"))
    if ssl_ca:
        creds["ssl_ca"] = ssl_ca
        creds["cacert"] = config_paths(CACERT)["target"]
    return creds


@when("exporter.do-reconfig")
def render_config():
    """Render the configuration for charm when all the interfaces are available."""
    hookenv.status_set("maintenance", "Updating configs")
    kv = unitdata.kv()
    creds = get_credentials()
    if not creds:
        hookenv.log("render_config: No credentials yet, skipping")
        hookenv.status_set(
            "blocked",
            "Waiting for credentials.  Please "
            "set os-credentials or add keystone relation",
        )
        return
    hookenv.log(
        "render_config: Got credentials for username={}".format(
            creds["credentials_username"]
        )
    )
    config = hookenv.config()
    ctx = config
    ctx["region"] = creds["region"]
    ctx["cache_file"] = os.path.join(VAR_SNAP_COMMON, creds["region"])
    ctx["cache_refresh_interval"] = config["cache-refresh-interval"]
    ctx["cpu_allocation_ratio"] = config["cpu-allocation-ratio"]
    ctx["ram_allocation_ratio"] = config["ram-allocation-ratio"]
    ctx["disk_allocation_ratio"] = config["disk-allocation-ratio"]
    ctx["schedulable_instance_size"] = config["schedulable-instance-size"].split(",")
    ctx["swift_hosts"] = kv.get("swift_hosts")
    ctx["use_nova_volumes"] = config["use_nova_volumes"]
    ctx["log_level"] = config["log_level"]
    if creds.get("ssl_ca"):
        hookenv.log("render_config: Using provided ssl_ca")
        # Use instead **kwargs due to PEP448 and trusty python 3.4.3
        render(
            source=config_paths(CACERT)["source"],
            target=config_paths(CACERT)["target"],
            context=creds,
        )
    # Use instead of **kwargs due to PEP448 and trusty python 3.4.3
    render(
        source=config_paths(CREDS)["source"],
        target=config_paths(CREDS)["target"],
        context=creds,
    )
    render(
        source=config_paths(MAIN)["source"],
        target=config_paths(MAIN)["target"],
        context=ctx,
    )
    remove_state("exporter.do-reconfig")
    set_state("exporter.do-restart")


def restart_service():
    if not host.service_running(SVC_NAME):
        hookenv.log("Starting {}...".format(SVC_NAME))
        host.service_start(SVC_NAME)
    else:
        hookenv.log("Restarting {}, config file changed...".format(SVC_NAME))
        host.service_restart(SVC_NAME)


@when("exporter.do-restart")
def do_restart():
    restart_service()
    hookenv.status_set("active", "Ready")
    remove_state("exporter.do-restart")


@when("prometheus-openstack-exporter-service.available")
def configure_exporter_service(exporter_service):
    config = hookenv.config()
    exporter_service.configure(config.get("port"))


@when("nrpe-external-master.available")
def update_nrpe_config(svc):
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    config = hookenv.config()
    port = config.get("port")
    extra_nrpe_args = config.get("extra-nrpe-args")
    # Note(aluria): check_http addresses LP#1829470 (/metrics takes too long, while / check takes ms
    # A final fix will land once LP#1829496 is fixed (we can't look now for "-s 'OpenStack Exporter'",
    # which is more explicit than "-s Exporter" body content)
    nrpe_setup.add_check(
        "prometheus_openstack_exporter_http",
        "Prometheus Openstack Exporter HTTP check",
        "check_http -I 127.0.0.1 -p {} -u / -s Exporter {}".format(
            port, extra_nrpe_args
        ),
    )
    nrpe_setup.write()


@when("prometheus-rules.available")
def render_prometheus_rules(prometheus_rules):
    # Send a list of rules for alerting to Prometheus
    config = hookenv.config()
    context = {
        "wait_time": config.get("wait_time"),
        "min_schedulable_instances": config.get("min_schedulable_instances"),
    }
    formatted_rules = []
    # This is a list, so we can add more templates in the future with minimal changes
    templates_dir = os.path.join(hookenv.charm_dir(), "templates")
    template_files = [
        "rule_schedulable_instances.j2",
    ]
    for template_file in template_files:
        with open(os.path.join(templates_dir, template_file), "r") as fd:
            formatted_rules.append(render_template(fd.read(), context))
    prometheus_rules.configure("\n".join(formatted_rules))


def render_template(template, context):
    tmpl = Template(template, lstrip_blocks=True, trim_blocks=True)
    return tmpl.render(**context)
