import base64
import os
import shutil
import tempfile
import unittest

from charmhelpers.core import unitdata
from charmhelpers.core.templating import render
import mock
from reactive import openstack_exporter
import yaml

# from charms.reactive import bus

os.environ["JUJU_UNIT_NAME"] = "prometheus-openstack-exporter"
os.environ["CHARM_DIR"] = "../../"


class SimpleConfigMock(dict):
    def __init__(self, *arg, **kw):
        super(SimpleConfigMock, self).__init__(*arg, **kw)
        self._changed_dict = {}
        self._changed_default = True

    def changed(self, key):
        return self._changed_dict.get(key, self._changed_default)

    def set_changed(self, changed_dict):
        self._changed_dict.update(changed_dict)


@mock.patch("os.chown")
@mock.patch("os.chmod")
@mock.patch("os.fchown")
@mock.patch("reactive.openstack_exporter.hookenv.status_set")
@mock.patch("reactive.openstack_exporter.configure_keystone_username")
@mock.patch("reactive.openstack_exporter.hookenv.unit_get")
@mock.patch("reactive.openstack_exporter.hookenv.config")
class TestOpenstackExporterContext(unittest.TestCase):
    def setUp(self):
        super(TestOpenstackExporterContext, self).setUp()
        self._init_tempdir_and_filenames()
        self._init_load_config_yml_defaults()
        self._init_override_render_for_nonroot()

    def _init_tempdir_and_filenames(self):
        self.dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.dir)
        self.maxDiff = None
        openstack_exporter.VAR_SNAP_ETC = self.dir
        # ugly hack, to avoid carrying global unitdata state across tests
        os.environ["UNIT_STATE_DB"] = os.path.join(self.dir, ".unit-state.db")
        unitdata._KV = None

    def _init_load_config_yml_defaults(self):
        # create def_config with parsed config.yaml defaults
        with open(os.path.join(os.environ["CHARM_DIR"], "config.yaml"), "r") as fh:
            config_yaml = yaml.safe_load(fh)
            self.def_config = SimpleConfigMock(
                {
                    k: v["default"]
                    for k, v in config_yaml["options"].items()
                    if v.get("default")
                }
            )

    def _init_override_render_for_nonroot(self):
        # need to override render() for 644 perms (ie allow overriding target
        # when running as non-root)
        def render_and_chmod(*args, **kwargs):
            render(*args, **kwargs)
            os.chmod(kwargs["target"], 0o664)

        self.mock_render = mock.patch(
            "reactive.openstack_exporter.render", side_effect=render_and_chmod
        )
        self.mock_render.start()
        self.addCleanup(self.mock_render.stop)

    def _config_file(self, key):
        return openstack_exporter.config_paths(key)["target"]

    def test_basic_config(self, _mock_hookenv_config, *args):
        exp_cert = "fortytwo"
        exp_cert64 = base64.encodestring(exp_cert.encode("utf-8")).decode()
        config = self.def_config
        config.update(
            {
                "os-credentials": yaml.dump(
                    {
                        "username": "foo",
                        "password": "b4r",
                        "region_name": "r1",
                        "auth_url": "http://127.0.0.1:5000/v2.0",
                    }
                ),
                "ssl_ca": exp_cert64,
            }
        )
        _mock_hookenv_config.return_value = config
        openstack_exporter.render_config()
        cfg_file = self._config_file(openstack_exporter.MAIN)
        yaml_content = yaml.safe_load(open(cfg_file))
        exp_keys = [
            "cache_file",
            "cache_refresh_interval",
            "cloud",
            "listen_port",
            "log_level",
        ]
        self.assertTrue(set(exp_keys).issubset(set(yaml_content.keys())))
        creds_file = self._config_file(openstack_exporter.CREDS)
        creds_content = open(creds_file).read().split("\n")
        exp_keys = [
            "OS_USERNAME",
            "OS_PASSWORD",
            "OS_REGION_NAME",
            "OS_AUTH_URL",
            "OS_TENANT_NAME",
            "OS_CACERT",
            "REQUESTS_CA_BUNDLE",
            "HOME",
        ]
        exp_keys = ["export {}".format(key) for key in exp_keys]
        act_creds = dict(
            [line.split("=") for line in creds_content if line.startswith("export")]
        )
        act_keys = act_creds.keys()
        self.assertSetEqual(set(exp_keys), set(act_keys))
        cacert_file = self._config_file(openstack_exporter.CACERT)
        self.assertEqual(exp_cert, open(cacert_file).read())

    @mock.patch("reactive.openstack_exporter.host.service_running", return_value=True)
    @mock.patch("reactive.openstack_exporter.host.service_restart")
    def test_restart(self, _mock_service_restart, *args):
        openstack_exporter.do_restart()
        _mock_service_restart.assert_called_once()

    @mock.patch("reactive.openstack_exporter.set_state")
    @mock.patch("reactive.openstack_exporter.data_changed")
    def test_reconfig_on_change(self, data_changed, set_state, *args):
        data_changed.return_value = True
        openstack_exporter.reconfig_on_change("foo", "{'bar': 'foo'}")
        data_changed.assert_called_once_with("foo", "{'bar': 'foo'}")
        set_state.assert_called_once_with("exporter.do-reconfig")

    @mock.patch("reactive.openstack_exporter.set_state")
    @mock.patch("reactive.openstack_exporter.data_changed")
    def test_reconfig_on_change_unchanged(self, data_changed, set_state, *args):
        data_changed.return_value = False
        openstack_exporter.reconfig_on_change("foo", "bar")
        set_state.assert_not_called()

    @mock.patch("reactive.openstack_exporter.render_template")
    def test_render_prometheus_rules(self, mock_render, mock_config, *args):
        prometheus_rules = mock.Mock()
        mock_render.return_value = "Test rendered template"
        openstack_exporter.render_prometheus_rules(prometheus_rules)
        prometheus_rules.configure.assert_called_with("Test rendered template")

    def test_render_template(self, mock_config, *args):
        context = {
            "wait_time": "5m",
            "min_schedulable_instances": 100,
        }
        with open(
            os.path.join(
                os.environ["CHARM_DIR"], "templates/rule_schedulable_instances.j2"
            ),
            "r",
        ) as template:
            result = openstack_exporter.render_template(template.read(), context)
        with open("rule_schedulable_instances", "r") as test_result:
            desired_result = test_result.read()
        assert result == desired_result
